
"""This module implements fuzzifiers used in mulearn.
"""

import numpy as np
from scipy.optimize import curve_fit
import warnings
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
import copy


def _safe_exp(r):
    with np.errstate(over="raise"):
        try:
            return np.exp(r)
        except FloatingPointError:
            return 1


class Fuzzifier:
    """Base class for fuzzifiers.

    The base class for fuzzifiers is Fuzzifier: it exposes a basic constructor
    which is called from the subclasses, and two methods `get_membership`
    (returning the membership function inferred from data) and `get_profile`
    computing information exploitable in order to visualize the fuzzifier
    in graphical form."""

    def __init__(self):
        """Create an instance of :class:`Fuzzifier`."""
        self.sq_radius_05 = None
        self.x_to_sq_dist = None
        self.r_to_mu = None

    def _get_r_to_mu(self):
        r"""Build membership function in feature space.

        Return a function that transforms the square distance between
        center of the learnt sphere and the image of a point in data
        space into a membership degree.

        **Note** This function is meant to be called internally by the
        `get_membership` method in the base `Fuzzifier` class.


        :returns: function -- function mapping square distance to membership.

        """

        check_is_fitted(self, ["r_to_mu"])
        return self.r_to_mu

    def get_membership(self):
        """Return the induced membership function.

        :raises: NotFittedError if `fit` has not been called
        :returns: function -- the induced membership function
        """
        r_to_mu = self._get_r_to_mu()
        return lambda x: r_to_mu(self.x_to_sq_dist(np.array(x)))

    def get_profile(self, X):
        r"""Return information about the learnt membership function profile.

        The profile of a membership function $\mu: X \rightarrow [0, 1]$ is
        intended here as the associated function $p: \mathbb R^+ \rightarrow
        [0, 1]$ still returning membership degrees, but considering its
        arguments in the feature space. More precisely, if `X` contains the
        values $x_1, \dots, x_n$, $R^2$ is the function mapping any point in
        data space into the squared distance between its image and the center
        $a$ of the learnt fuzzy set in feature space, the function
        `get_profile` computes the following information about $p$:

        * a list $r_\mathrm{data} = [ R^2(x_i), i = 1, \dots, n]$ containing
          the distances between the images of the points in `X` and $a$;
        * a list $\tilde{r}_\mathrm{data}$ containing 200 possible
          distances between $a$ and the image of a point in data space, evenly
          distributed between $0$ and $\max r_{\mathrm{data}}$;
        * a list $e = [\hat\mu(r_i), r_i \in \tilde{r}_{\mathrm{data}}]$
          gathering the profile values for each element in
          $\tilde{r}_{\mathrm{data}}$.

        This information can be used in order to graphically show the
        membership profile, which is always plottable, whereas the membership
        function isn't mostly of the time (unless the elements in `X` are
        either one- or bidimensional vectors).

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :returns: list -- $[r_{\mathrm{data}}, \tilde{r}_\mathrm{data}, e]$.

        """
        rdata = list(map(self.x_to_sq_dist, X))
        rdata_synth = np.linspace(0, max(rdata) * 1.1, 200)
        estimate = list(map(self._get_r_to_mu(), rdata_synth))
        return [rdata, rdata_synth, estimate]

    def __str__(self):
        """Return the string representation of a fuzzifier."""
        return self.__repr__()

    def __eq__(self, other):
        """Check fuzzifier equality w.r.t. other objects."""
        return type(self) == type(other)

    def __ne__(self, other):
        """Check fuzzifier inequality w.r.t. other objects."""
        return not self == other

    def __hash__(self):
        """Generate hashcode for a fuzzifier."""
        return hash(self.__repr__())

    @staticmethod
    def __nonzero__():
        """Check if a fuzzifier is non-null."""
        return True

    def __getstate__(self):
        """Return a serializable description of the fuzzifier."""
        d = copy.deepcopy(self.__dict__)
        if 'x_to_sq_dist' in d:
            del d['x_to_sq_dist']
        return d

    def __setstate__(self, d):
        """Ensure fuzzifier consistency after deserialization."""
        self.__dict__ = d


class CrispFuzzifier(Fuzzifier):
    """Crisp fuzzifier.

    Fuzzifier corresponding to a crisp (classical) set: membership is always
    equal to either $0$ or $1$."""

    default_profile = "fixed"

    def __init__(self, profile=default_profile):
        r"""Create an instance of :class:`CrispFuzzifier`.

        :param profile: method to be used in order to build the fuzzifier
          profile: `'fixed'` relies on the radius of the sphere defining
          the fuzzy set core, while `'infer'` fits a generic threshold
          function on the provided examples.
        :type profile: str
        """
        super().__init__()

        self.profile = profile

        self.name = 'Crisp'
        self.latex_name = '$\\hat\\mu_{\\text{crisp}}$'

    def fit(self, X, y):
        r"""Fit the fuzzifier on training data.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :param y: membership degrees of the values in `X`.
        :type y: vector of floats having the same length of `X`
        :raises: ValueError if self.profile is not set either to `'fixed'` or
          to `'infer'`.

        The fitting process is done considering a threshold-based membership
        function, in turn corresponding to a threshold-based profile of the
        form

        .. math::
          p(r) = \begin{cases} 1 & \text{if $r \leq r_\text{crisp}$,} \\
                               0 & \text{otherwise.} \end{cases}

        The threshold $r_\text{crisp}$ is set to the learnt square radius of
        the sphere when the `profile` attribute of the class have been set to
        `'fixed'`, and induced via interpolation of `X` and `y` attributes when
        it is has been set to `'infer'`.
        """
        check_array(X)
        check_X_y(X, y)

        if self.profile == "fixed":
            self.r_to_mu = lambda r: 1 if r <= self.sq_radius_05 else 0

        elif self.profile == "infer":
            R = np.fromiter(map(self.x_to_sq_dist, X), dtype=float)

            def r_to_mu(r, sq_radius_05):
                result = np.ones(len(r))
                result[r > sq_radius_05] = 0
                return result

            p_opt, _ = curve_fit(r_to_mu, R, y,
                                 bounds=((0,), (np.inf,)))

            if p_opt[0] < 0:
                raise ValueError("Profile fit returned a negative parameter")

            self.r_to_mu = lambda r: r_to_mu([r], *p_opt)[0]
        else:
            raise ValueError("'profile' parameter should either be equal to "
                             f"'fixed' or 'infer' (provided: {self.profile})")

    def __repr__(self):
        """Return the python representation of the fuzzifier."""
        if self.profile != self.default_profile:
            return f"CrispFuzzifier(profile={self.profile})"
        else:
            return "CrispFuzzifier()"


class LinearFuzzifier(Fuzzifier):
    """Crisp fuzzifier.

    Fuzzifier corresponding to a fuzzy set whose membership in feature space
    linearly decreases from 1 to 0."""

    default_profile = "fixed"

    def __init__(self, profile=default_profile):
        r"""Create an instance of :class:`LinearFuzzifier`.

        :param profile: method to be used in order to build the fuzzifier
          profile: `'fixed'` relies on the radius of the sphere defining
          the fuzzy set core, while `'infer'` fits the profile function on the
          provided examples.
        :type profile: str
        """
        super().__init__()

        self.profile = profile

        self.name = 'Linear'
        self.latex_name = '$\\hat\\mu_{\\text{lin}}$'

    def fit(self, X, y):
        r"""Fit the fuzzifier on training data.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :param y: membership degrees of the values in `X`.
        :type y: vector of floats having the same length of `X`
        :raises: ValueError if self.profile is not set either to `'fixed'` or
          to `'infer'`.

        The fitting process is done considering a membership function
        linearly decreasing from $1$ to $0$, in turn corresponding to a profile
        having the general form

        .. math::
          p(r) = \begin{cases} 1 & \text{if $r \leq r_1$,} \\
                               l(r) & \text{if $r_1 < r \leq r_0$,} \\
                               0 & \text{otherwise.} \end{cases}

        The free parameters are chosen in order to guarantee continuity;
        moreover, when the `profile` attribute of the class have been set to
        `'fixed'` the membership profile will be equal to 0.5 when $r$ is equal
        to the learnt square radius of the sphere, and induced via
        interpolation of `X` and `y` when it is has been set to `'infer'`.
        """
        check_array(X)
        check_X_y(X, y)
        R = np.fromiter(map(self.x_to_sq_dist, X), dtype=float)

        sq_radius_1_guess = np.median([self.x_to_sq_dist(x)
                                       for x, mu in zip(X, y) if mu >= 0.99])

        if self.profile == 'fixed':
            def r_to_mu(R_arg, sq_radius_1):
                return [np.clip(1 - 0.5 *
                                (r - sq_radius_1) /
                                (self.sq_radius_05 - sq_radius_1),
                                0, 1)
                        for r in R_arg]

            p_opt, _ = curve_fit(r_to_mu, R, y,
                                 p0=(sq_radius_1_guess,),
                                 bounds=((0,), (np.inf,)))

        elif self.profile == 'infer':

            def r_to_mu(R_arg, sq_radius_1, sq_radius_0):
                return [np.clip(1 - (sq_radius_1 - r) /
                                (sq_radius_1 - sq_radius_0), 0, 1)
                        for r in R_arg]

            p_opt, _ = curve_fit(r_to_mu, R, y,
                                 p0=(sq_radius_1_guess, 10 * self.sq_radius_05),
                                 bounds=((0, 0), (np.inf, np.inf,)))
        else:
            raise ValueError("'profile' parameter should be equal to "
                             "'fixed' or 'infer' (provided value: {profile})")
        if min(p_opt) < 0:
            raise ValueError('Profile fitting returned a negative parameter')

        self.r_to_mu = lambda r: r_to_mu([r], *p_opt)[0]

    def __repr__(self):
        """Return the python representation of the fuzzifier."""
        if self.profile != self.default_profile:
            return f"LinearFuzzifier(profile={self.profile})"
        else:
            return "LinearFuzzifier()"


class ExponentialFuzzifier(Fuzzifier):
    """Exponential fuzzifier.

    Fuzzifier corresponding to a fuzzy set whose membership in feature space
    exponentially decreases from 1 to 0."""

    default_profile = "fixed"
    default_alpha = -1

    def __init__(self, profile=default_profile, alpha=default_alpha):
        r"""Create an instance of :class:`ExponentialFuzzifier`.

        :param profile: method to be used in order to build the fuzzifier
          profile: `'fixed'` relies on the radius of the sphere defining
          the fuzzy set core, `'infer'` fits the profile function on the
          provided examples, and `'alpha'` allows for manually setting the
          exponential decay via the `alpha` parameter.
        :type profile: str
        :param alpha: fixed exponential decay of the fuzzifier.
        :type alpha: float
        """

        super().__init__()

        self.profile = profile
        self.alpha = alpha

        self.name = "Exponential"
        self.latex_name = r"$\hat\mu_{\text{exp}}$"

    def fit(self, X, y):
        r"""Fit the fuzzifier on training data.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :param y: membership degrees of the values in `X`.
        :type y: vector of floats having the same length of `X`
        :raises: ValueError if self.profile is not set either to `'fixed'`,
          `'infer'`, or `'alpha'`.

        In this fuzzifier, the function that transforms the square distance
        between the center of the learnt sphere and the image of a point in
        the original space into a membership degree has the form

        .. math::
          \mu(r) = \begin{cases}  1    & \text{if $r \leq r_1$,} \\
                                  e(r) & \text{otherwise,}
                   \end{cases}

        where $e$ is an exponential function decreasing from 1 to 0. The
        shape of this function is chosen so that the membership profile will be
        equal to 0.5 when $r$ is equal to the learnt square radius of the
        sphere, and induced via interpolation of `X` and `y` when it is has
        been set to `'infer'`; finally, when the parameter is set to `'alpha'`
        the exponential decay of $e$ is manually set via the `alpha` parameter
        of the class constructor.
        """
        check_array(X)
        check_X_y(X, y)

        if self.alpha > 0 and self.profile != "alpha":
            raise ValueError(f"'alpha' value is specified, but 'profile' "
                             f"is set to '{self.profile}'")

        if self.profile == "alpha":
            if self.alpha < 0 or self.alpha > 1:
                raise ValueError("alpha must be set to a float between 0 and 1 "
                                 "when 'profile' is 'alpha'")

        r_1_guess = np.median([self.x_to_sq_dist(x)
                               for x, mu in zip(X, y) if mu >= 0.99])

        s_guess = (self.sq_radius_05 - r_1_guess) / np.log(2)

        R = np.fromiter(map(self.x_to_sq_dist, X), dtype=float)

        if self.profile == "fixed":
            def r_to_mu(R_data, sq_radius_1):
                return [np.clip(_safe_exp(-(r - sq_radius_1) /
                                          (self.sq_radius_05 - sq_radius_1)
                                          * np.log(2)), 0, 1) for r in R_data]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                p_opt, _ = curve_fit(r_to_mu, R, y, p0=(r_1_guess,),
                                     maxfev=2000, bounds=((0,), (np.inf,)))
            self.r_to_mu = lambda r: r_to_mu([r], *p_opt)[0]

        elif self.profile == "infer":
            def r_to_mu(R_data, r_1, s):
                return [np.clip(_safe_exp(-(r - r_1) / s), 0, 1)
                        for r in R_data]

            p_opt, _ = curve_fit(r_to_mu, R, y, p0=(r_1_guess, s_guess),
                                 # bounds=((0, 0), (np.inf, np.inf)),
                                 maxfev=2000)

            self.r_to_mu = lambda r: r_to_mu([r], *p_opt)[0]

        elif self.profile == "alpha":
            r_sample = map(self.x_to_sq_dist, X)

            q = np.percentile([s - self.sq_radius_05 for s in r_sample
                               if s > self.sq_radius_05], 100 * self.alpha)

            def r_to_mu(R_data, sq_radius_1):
                return [np.clip(_safe_exp(np.log(self.alpha) /
                                          q * (r - sq_radius_1)), 0, 1)
                        for r in R_data]

            p_opt, _ = curve_fit(r_to_mu, R, y, p0=(r_1_guess,),
                                 bounds=((0,), (np.inf,)))
            self.r_to_mu = lambda r: r_to_mu([r], *p_opt)[0]
        else:
            raise ValueError("'profile' parameter should be equal to "
                             "'infer', 'fixed' or 'alpha' "
                             f"(provided value: {self.profile})")

    def __repr__(self):
        obj_repr = "ExponentialFuzzifier("
        if self.profile != self.default_profile:
            obj_repr += f", profile={self.profile}"
        if self.alpha != self.default_alpha:
            obj_repr += f", alpha={self.alpha}"
        if obj_repr.endswith(", "):
            return obj_repr + ")"
        else:
            return "ExponentialFuzzifier()"


class QuantileConstantPiecewiseFuzzifier(Fuzzifier):
    """Quantile-based constant piecewise fuzzifier.

    Fuzzifier corresponding to a fuzzy set with a piecewise constant membership
    function, whose steps are defined according to the quartiles of the squared
    distances between images of points and center of the learnt sphere."""

    def __init__(self):
        r"""Create an instance of :class:`QuantileConstantPiecewiseFuzzifier`"""

        super().__init__()

        self.name = 'QuantileConstPiecewise'
        self.latex_name = '$\\hat\\mu_{\\text{q\\_const}}$'

    def fit(self, X, y):
        """Fit the fuzzifier on training data.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :param y: membership degrees of the values in `X`.
        :type y: vector of floats having the same length of `X`

        The piecewise membership function is built so that its steps are chosen
        according to the quartiles of square distances between images of the
        points in `X` center of the learnt sphere.
        """
        check_array(X)
        check_X_y(X, y)

        R = np.fromiter(map(self.x_to_sq_dist, X), dtype=float)

        sq_radius_1 = np.median([self.x_to_sq_dist(x)
                                 for x, mu in zip(X, y) if mu >= 0.99])
        external_dist = [r - sq_radius_1
                         for r in R if r > sq_radius_1]

        if external_dist:
            m = np.median(external_dist)
            q1 = np.percentile(external_dist, 25)
            q3 = np.percentile(external_dist, 75)
        else:
            m = q1 = q3 = 0

        def r_to_mu(r):
            return 1 if r <= sq_radius_1 \
                else 0.75 if r <= sq_radius_1 + q1 \
                else 0.5 if r <= sq_radius_1 + m \
                else 0.25 if r <= sq_radius_1 + q3 \
                else 0

        self.r_to_mu = r_to_mu

    def __repr__(self):
        return "QuantileConstantPiecewiseFuzzifier()"


class QuantileLinearPiecewiseFuzzifier(Fuzzifier):
    """Quantile-based linear piecewise fuzzifier.

    Fuzzifier corresponding to a fuzzy set with a piecewise linear membership
    function, whose steps are defined according to the quartiles of the squared
    distances between images of points and center of the learnt sphere."""

    def __init__(self):
        r"""Create an instance of :class:`QuantileLinearPiecewiseFuzzifier`."""

        super().__init__()

        self.name = 'QuantileLinPiecewise'
        self.latex_name = '$\\hat\\mu_{\\text{q\\_lin}}$'

    def fit(self, X, y):
        """Fit the fuzzifier on training data.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :param y: membership degrees of the values in `X`.
        :type y: vector of floats having the same length of `X`

        The piecewise membership function is built so that its steps are chosen
        according to the quartiles of square distances between images of the
        points in `X` center of the learnt sphere.
        """
        check_array(X)
        check_X_y(X, y)

        R = np.fromiter(map(self.x_to_sq_dist, X), dtype=float)

        sq_radius_1 = np.median([self.x_to_sq_dist(x)
                                 for x, mu in zip(X, y) if mu >= 0.99])
        external_dist = [r - sq_radius_1
                         for r in R if r > sq_radius_1]

        if external_dist:
            m = np.median(external_dist)
            q1 = np.percentile(external_dist, 25)
            q3 = np.percentile(external_dist, 75)
            mx = np.max(external_dist)
        else:
            m = q1 = q3 = mx = 0

        def r_to_mu(r):
            ssd = sq_radius_1
            return 1 if r <= ssd \
                else (-r + ssd) / (4 * q1) + 1 if r <= ssd + q1 \
                else (-r + ssd + q1) / (4 * (m - q1)) + 3 / 4 if r <= ssd + m \
                else (-r + ssd + m) / (4 * (q3 - m)) + 1 / 2 if r <= ssd + q3 \
                else (-r + ssd + q3) / (4 * (mx - q3)) + 1 / 4 if r <= ssd + mx\
                else 0

        self.r_to_mu = r_to_mu

    def __repr__(self):
        return "QuantileLinearPiecewiseFuzzifier()"



"""Implementation of optimization procedures.

This module module contains the implementations of the optimization processes
behind fuzzy inference.

Once loaded, the module preliminarily verifies that some libraries are
installed (notably, Gurobi and TensorFlow), emitting a warning otherwise.
Note that at least one of these libraries is needed in order to solve the
optimization problems involved in the fuzzy inference process.

The module also checks the availability of tqdm, which is used in order to
graphically depict the progress of some learning processes using a progress
bar. However, this package is not strictly needed: if it is not installed,
the above mentioned progress bars will not be displayed.
"""

import numpy as np
import itertools as it
from collections.abc import Iterable
import logging

import mulearn.kernel as kernel

logger = logging.getLogger(__name__)

try:
    from gurobipy import LinExpr, GRB, Model, Env, QuadExpr, GurobiError

    gurobi_ok = True
except ModuleNotFoundError:
    logger.warning('gurobi not available')
    gurobi_ok = False

try:
    import os

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    logging.getLogger('tensorflow').setLevel(logging.FATAL)
    import tensorflow as tf

    tensorflow_ok = True
    from tensorflow.keras.optimizers import Adam

    logging.getLogger('tensorflow').setLevel(logging.ERROR)
except ModuleNotFoundError:
    logger.warning('tensorflow not available')
    tensorflow_ok = False

try:
    import tqdm

    tqdm_ok = True
except ModuleNotFoundError:
    logger.warning('tqdm not available')
    tqdm_ok = False


class Solver:
    """Abstract solver for optimization problems.

    The base class for solvers is :class:`Solver`: it exposes a method
    `solve` which delegates the numerical optimization process to an abstract
    method `solve_problem` and subsequently clips the results to the boundaries
    of the feasible region.
    """

    def solve_problem(self, *args):
        pass

    def solve(self, xs, mus, c, k):
        """Solve optimization phase.

        Build and solve the constrained optimization problem on the basis
        of the fuzzy learning procedure.

        :param xs: Objects in training set.
        :type xs: iterable
        :param mus: Membership values for the objects in `xs`.
        :type mus: iterable
        :param c: constant managing the trade-off in joint radius/error
          optimization.
        :type c: float
        :param k: Kernel function to be used.
        :type k: :class:`mulearn.kernel.Kernel`
        :raises: ValueError if c is non-positive or if xs and mus have
          different lengths.
        :returns: `list` -- optimal values for the independent variables
          of the problem."""
        if c <= 0:
            raise ValueError('c should be positive')

        mus = np.array(mus)
        chis = self.solve_problem(xs, mus, c, k)

        chis_opt = [np.clip(ch, l, u)
                    for ch, l, u in zip(chis, -c * (1 - mus), c * mus)] # noqa

        return chis_opt


class GurobiSolver(Solver):
    """Solver based on gurobi.

    Using this class requires that gurobi is installed and activated
    with a software key. The library is available at no cost for academic
    purposes (see
    https://www.gurobi.com/downloads/end-user-license-agreement-academic/).
    Alongside the library, also its interface to python should be installed,
    via the gurobipy package.
    """

    default_values = {"time_limit": 10 * 60,
                      "adjustment": 0,
                      "initial_values": None}

    def __init__(self, time_limit=default_values['time_limit'],
                 adjustment=default_values['adjustment'],
                 initial_values=default_values['initial_values']):
        """
        Build an object of type GurobiSolver.

        :param time_limit: Maximum time (in seconds) before stopping iterative
          optimization, defaults to 10*60.
        :type time_limit: int
        :param adjustment: Adjustment value to be used with non-PSD matrices,
          defaults to 0. Specifying `'auto'` instead than a numeric value
          will automatically trigger the optimal adjustment if needed.
        :type adjustment: float or `'auto'`
        :param initial_values: Initial values for variables of the optimization
          problem, defaults to None.
        :type initial_values: iterable of floats or None
        """
        self.time_limit = time_limit
        self.adjustment = adjustment
        self.initial_values = initial_values

    def solve_problem(self, xs, mus, c, k):
        """Optimize via gurobi.

        Build and solve the constrained optimization problem at the basis
        of the fuzzy learning procedure using the gurobi API.

        :param xs: objects in training set.
        :type xs: iterable
        :param mus: membership values for the objects in `xs`.
        :type mus: iterable
        :param c: constant managing the trade-off in joint radius/error
          optimization.
        :type c: float
        :param k: kernel function to be used.
        :type k: :class:`mulearn.kernel.Kernel`
        :raises: ValueError if optimization fails or if gurobi is not installed
        :returns: list -- optimal values for the independent variables of the
          problem.
        """
        if not gurobi_ok:
            raise ValueError('gurobi not available')

        m = len(xs)

        with Env(empty=True) as env:
            env.setParam('OutputFlag', 0)
            env.start()
            with Model('mulearn', env=env) as model:
                model.setParam('OutputFlag', 0)
                model.setParam('TimeLimit', self.time_limit)

                for i in range(m):
                    if c < np.inf:
                        model.addVar(name=f'chi_{i}',
                                     lb=-c * (1 - mus[i]), ub=c * mus[i],
                                     vtype=GRB.CONTINUOUS)
                    else:
                        model.addVar(name=f'chi_{i}', vtype=GRB.CONTINUOUS)

                model.update()
                chis = model.getVars()

                if self.initial_values is not None:
                    for c, i in zip(chis, self.initial_values):
                        c.start = i

                obj = QuadExpr()

                for i, j in it.product(range(m), range(m)):
                    obj.add(chis[i] * chis[j], k.compute(xs[i], xs[j]))

                for i in range(m):
                    obj.add(-1 * chis[i] * k.compute(xs[i], xs[i]))

                if self.adjustment and self.adjustment != 'auto':
                    for i in range(m):
                        obj.add(self.adjustment * chis[i] * chis[i])

                model.setObjective(obj, GRB.MINIMIZE)

                constEqual = LinExpr()
                constEqual.add(sum(chis), 1.0)

                model.addConstr(constEqual, GRB.EQUAL, 1)

                try:
                    model.optimize()
                except GurobiError as e:
                    print(e.message)
                    if self.adjustment == 'auto':
                        s = e.message
                        a = float(s[s.find(' of ') + 4:s.find(' would')])
                        logger.warning('non-diagonal Gram matrix, '
                                       f'retrying with adjustment {a}')
                        for i in range(m):
                            obj.add(a * chis[i] * chis[i])
                        model.setObjective(obj, GRB.MINIMIZE)

                        model.optimize()
                    else:
                        raise e

                if model.Status != GRB.OPTIMAL:
                    raise ValueError('optimal solution not found!')

                return [ch.x for ch in chis]

    def __repr__(self):
        obj_repr = f"GurobiSolver("

        for a in ('time_limit', 'adjustment', 'initial_values'):
            if self.__getattribute__(a) != self.default_values[a]:
                obj_repr += f", {a}={self.default_values[a]}"
        return obj_repr + ")"


class TensorFlowSolver(Solver):
    """Solver based on TensorFlow.

    Using this class requires that TensorFlow 2.X is installed."""

    default_values = {"initial_values": "random",
                      "init_bound": 0.1,
                      "n_iter": 100,
                      "optimizer": Adam(learning_rate=1e-4)
                                   if tensorflow_ok else None,  # noqa
                      "tracker": tqdm.trange if tqdm_ok else range,
                      "penalization": 10}

    def __init__(self,
                 initial_values=default_values["initial_values"],
                 init_bound=default_values["init_bound"],
                 n_iter=default_values["n_iter"],
                 optimizer=default_values["optimizer"],
                 tracker=default_values["tracker"],
                 penalization=default_values["penalization"]):
        """Build an object of type TensorFlowSolver.

        :param initial_values: values to be used for initializing the
          independent  variables, either randomly (if set to `'random'`)` or
          seting them to a given sequence of initial values (if set to an
          iterable of floats), defaults to `'random'`.
        :type initial_values: `str` or iterable of `float`
        :param init_bound: Absolute value of the extremes of the interval used
          for random initialization of independent variables, defaults to 0.1.
        :type init_bound: `float`
        :param n_iter: Number of iterations of the optimization process,
          defaults to 100.
        :type n_iter: `int`
        :param optimizer: Optimization algorithm to be used, defaults to Adam
          with learning rate=1e-4 if tensorflow is available, to None otherwise.
        :type optimizer: :class:`tf.keras.optimizers.Optimizer`
        :param tracker: Tool to graphically depict the optimization progress,
          defaults to `tqdm.trange` if tqdm is available, to `range` (i.e., no
          graphical tool) otherwise.
        :type tracker: `object`
        :param penalization: Lagrange penalization for the equality constraint
          in the original problem, defaults to 10.
        :type penalization: `float`
        """
        self.init_bound = init_bound
        self.initial_values = initial_values
        self.n_iter = n_iter
        self.optimizer = optimizer
        self.tracker = tracker
        self.penalization = penalization

    def solve_problem(self, xs, mus, c, k):
        """Optimize via TensorFlow.

        Build and solve the constrained optimization problem on the basis
        of the fuzzy learning procedure using the TensorFlow API.

        :param xs: objects in training set.
        :type xs: iterable
        :param mus: membership values for the objects in `xs`.
        :type mus: iterable
        :param c: constant managing the trade-off in joint radius/error
          optimization.
        :type c: float
        :param k: kernel function to be used.
        :type k: :class:`mulearn.kernel.Kernel`
        :raises: ValueError if optimization fails or if TensorFlow is not
          installed
        :returns: list -- optimal values for the independent variables of the
          problem.
        """
        if not tensorflow_ok:
            raise ValueError('tensorflow not available')

        m = len(xs)

        if self.initial_values == 'random':
            chis = [tf.Variable(ch, name=f'chi_{i}',
                                trainable=True, dtype=tf.float32)
                    for i, ch in enumerate(np.random.uniform(-self.init_bound,
                                                             self.init_bound,
                                                             m))]
        elif isinstance(self.initial_values, Iterable):
            chis = [tf.Variable(ch, name=f'chi_{i}',
                                trainable=True, dtype=tf.float32)
                    for i, ch in enumerate(self.initial_values)]
        else:
            raise ValueError("`initial_values` should either be set to "
                             "'random' or to a list of initial values.")

        if type(k) is kernel.PrecomputedKernel:
            gram = k.kernel_computations
        else:
            gram = np.array([[k.compute(x1, x2) for x1 in xs] for x2 in xs])

        def obj():
            kernels = tf.constant(gram, dtype='float32')

            v = tf.tensordot(tf.linalg.matvec(kernels, chis), chis, axes=1)
            v -= tf.tensordot(chis,
                              [k.compute(x_i, x_i) for x_i in xs], axes=1)

            v += self.penalization * tf.math.maximum(0, 1 - sum(chis))
            v += self.penalization * tf.math.maximum(0, sum(chis) - 1)

            if c < np.inf:
                for chi, mu in zip(chis, mus):
                    v += self.penalization * tf.math.maximum(0, chi - c * mu)
                    v += self.penalization *\
                         tf.math.maximum(0, c * (1 - mu) - chi) # noqa

            return v

        for _ in self.tracker(self.n_iter):
            self.optimizer.minimize(obj, var_list=chis)

        return [ch.numpy() for ch in chis]

    def __repr__(self):
        obj_repr = f"TensorFlowSolver("

        for a in ("initial_values", "init_bound", "n_iter",
                  "optimizer", "tracker", "penalization"):
            if self.__getattribute__(a) != self.default_values[a]:
                obj_repr += f", {a}={self.default_values[a]}"
        return obj_repr + ")"



"""Kernel implementation.

This module implements the kernel used in mulearn.
"""

import numpy as np


class Kernel:
    """Base kernel class."""

    def __init__(self):
        """Create an instance of :class:`Kernel`."""
        self.precomputed = False
        self.kernel_computations = None

    def compute(self, arg_1, arg_2):
        """Compute the kernel value, given two arguments.

        :param arg_1: First kernel argument.
        :type arg_1: Object
        :param arg_2: Second kernel argument.
        :type arg_2: Object
        :raises: NotImplementedError (:class:`Kernel` is abstract)
        :returns: `float` -- kernel value.
        """
        raise NotImplementedError(
            'The base class does not implement the `compute` method')

    def __str__(self):
        """Return the string representation of a kernel."""
        return self.__repr__()

    def __eq__(self, other):
        """Check kernel equality w.r.t. other objects."""
        return type(self) == type(other)

    def __ne__(self, other):
        """Check kernel inequality w.r.t. other objects."""
        return not self == other

    @staticmethod
    def __nonzero__():
        """Check if a kernel is non-null."""
        return True

    def __hash__(self):
        """Generate hashcode for a kernel."""
        return hash(self.__repr__())

    @classmethod
    def get_default(cls):
        """Return the default kernel.

        :returns: `LinearKernel()` -- the default kernel.
        """
        return LinearKernel()


class LinearKernel(Kernel):
    """Linear kernel class."""

    def compute(self, arg_1, arg_2):
        r"""
        Compute the kernel value.

        The value $k(x_1, x_2)$ of a linear kernel is equal to the dot product
        $x_1 \cdot x_2$, that is to $\sum_{i=1}^n (x_1)_i (x_2)_i$, $n$ being
        the common dimension of $x_1$ and $x_2$.

        :param arg_1: First kernel argument.
        :type arg_1: iterable of `float`
        :param arg_2: Second kernel argument.
        :type arg_2: iterable of `float`
        :returns: `float` -- kernel value.
        """
        return float(np.dot(arg_1, arg_2))

    def __repr__(self):
        """Return the python representation of the kernel."""
        return 'LinearKernel()'


class PolynomialKernel(Kernel):
    """Polynomial kernel class."""

    def __init__(self, degree):
        r"""Create an instance of `PolynomialKernel`.

        :param degree: degree of the polynomial kernel.
        :type degree: `int`
        :raises: ValueError if `degree` is not an integer or if it has a
          negative value.
        """
        super().__init__()
        if degree > 0 and isinstance(degree, int):
            self.degree = degree
        else:
            raise ValueError(f"{degree} is not usable as a polynomial degree")

    def compute(self, arg_1, arg_2):
        r"""
        Compute the kernel value.

        The value $k(x_1, x_2)$ of a polynomial kernel is equal to the
        quantity $(x_1 \cdot x_2 + 1)^d$, $d$ being the polynomial degree of
        the kernel.

        :param arg_1: First kernel argument.
        :type arg_1: iterable of `float`
        :param arg_2: Second kernel argument.
        :type arg_2: iterable of `float`
        :returns: `float` -- kernel value.
        """
        return float((np.dot(arg_1, arg_2) + 1) ** self.degree)

    def __repr__(self):
        """Return the python representation of the kernel."""
        return f"PolynomialKernel({self.degree})"


class HomogeneousPolynomialKernel(PolynomialKernel):
    """Homogeneous polynomial kernel class."""

    def __init__(self, degree):
        r"""Create an instance of `HomogeneousPolynomialKernel`.

        :param degree: degree of the polynomial kernel.
        :type degree: `int`
        :raises: ValueError if `degree` is not an integer or if it has a
          negative value.
        """
        super().__init__(degree)

    def compute(self, arg_1, arg_2):
        r"""Compute the kernel value.

        The value $k(x_1, x_2)$ of a homogeneous polynomial kernel is
        intended as the quantity $(x_1 \cdot x_2)^d$, $d$ being the polynomial
        degree of the kernel.

        :param arg_1: First kernel argument.
        :type arg_1: iterable of `float`
        :param arg_2: Second kernel argument.
        :type arg_2: iterable of `float`
        :returns: `float` -- kernel value.
        """
        return float(np.dot(arg_1, arg_2) ** self.degree)

    def __repr__(self):
        """Return the python representation of the kernel."""
        return f"HomogeneousPolynomialKernel({self.degree})"


class GaussianKernel(Kernel):
    """Gaussian kernel class."""

    default_sigma = 1

    def __init__(self, sigma=default_sigma):
        r"""Create an instance of `GaussianKernel`.

        :param sigma: gaussian standard deviation, defaults to 1.
        :type sigma: `float`
        :raises: ValueError if `sigma` has a negative value.
        """
        super().__init__()
        if sigma > 0:
            self.sigma = sigma
        else:
            raise ValueError(f'{sigma} is not usable '
                             'as a gaussian standard deviation')

    def compute(self, arg_1, arg_2):
        r"""Compute the kernel value.

        The value $k(x_1, x_2)$ of a gaussian kernel is intended as the
        quantity $\mathrm e^{-\frac{||x_1 - x_2||^2}{2 \sigma^2}}$, $\sigma$
        being the kernel standard deviation.

        :param arg_1: First kernel argument.
        :type arg_1: iterable of `float`
        :param arg_2: Second kernel argument.
        :type arg_2: iterable of `float`
        :returns: `float` -- kernel value.
        """
        diff = np.linalg.norm(np.array(arg_1) - np.array(arg_2)) ** 2
        return float(np.exp(-1. * diff / (2 * self.sigma ** 2)))

    def __repr__(self):
        """Return the python representation of the kernel."""
        obj_repr = "GaussianKernel("
        if self.sigma != self.default_sigma:
            obj_repr += f"sigma={self.sigma}"
        obj_repr += ")"
        return obj_repr


class HyperbolicKernel(Kernel):
    """Hyperbolic kernel class."""

    default_scale = 1
    default_offset = 0

    def __init__(self, scale=default_scale, offset=default_offset):
        r"""Create an instance of `HyperbolicKernel`.

        :param scale: scale constant, defaults to 1.
        :type scale: `float`
        :param offset: offset constant, defaults to 0.
        :type offset: `float`
        """
        super().__init__()
        self.scale = scale
        self.offset = offset

    def compute(self, arg_1, arg_2):
        r"""Compute the kernel value.

        The value $k(x_1, x_2)$ of a hyperbolic kernel is intended as the
        quantity $\tanh(\alpha x_1 \cdot x_2 + \beta)$, $\alpha$ and $\beta$
        being the scale and offset parameters, respectively.

        :param arg_1: First kernel argument.
        :type arg_1: iterable of `float`
        :param arg_2: Second kernel argument.
        :type arg_2: iterable of `float`
        :returns: `float` -- kernel value.
        """
        dot_orig = np.dot(np.array(arg_1), np.array(arg_2))
        return float(np.tanh(self.scale * dot_orig + self.offset))

    def __repr__(self):
        """Return the python representation of the kernel."""
        obj_repr = "HyperbolicKernel("
        if self.scale != self.default_scale:
            obj_repr += f"scale={self.scale}, "
        if self.offset != self.default_offset:
            obj_repr += f"offset={self.offset}, "

        if obj_repr.endswith(", "):
            return obj_repr[:-2] + ")"
        else:
            return "HyperbolicKernel()"


class PrecomputedKernel(Kernel):
    """Precomputed kernel class."""

    def __init__(self, kernel_computations):
        r"""Create an instance of `PrecomputedKernel`.

        :param kernel_computations: kernel computations.
        :type kernel_computations: square matrix of float elements
        :raises: ValueError if `kernel_computations` is not a square
          bidimensional array.
        """
        super().__init__()
        self.precomputed = True
        try:
            (rows, columns) = np.array(kernel_computations).shape
        except ValueError:
            raise ValueError('The supplied matrix is not array-like ')

        if rows != columns:
            raise ValueError('The supplied matrix is not square')

        self.kernel_computations = kernel_computations

    def compute(self, arg_1, arg_2):
        r"""Compute the kernel value.

        The value of a precomputed kernel is retrieved according to the indices
        of the corresponding objects. Note that each index should be enclosed
        within an iterable in order to be compatible with sklearn.

        :param arg_1: First kernel argument.
        :type arg_1: iterable of `float`
        :param arg_2: Second kernel argument.
        :type arg_2: iterable of `float`
        :returns: `float` -- kernel value.
        """
        return float(self.kernel_computations[arg_1[0]][arg_2[0]])

    def __repr__(self):
        """Return the python representation of the kernel."""
        return f"PrecomputedKernel({self.kernel_computations})"




import numpy as np
import copy

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils import check_random_state
from sklearn.exceptions import NotFittedError

from mulearn.kernel import GaussianKernel
from mulearn.optimization import GurobiSolver
from mulearn.fuzzifier import ExponentialFuzzifier

import logging

import warnings
from scipy.optimize import OptimizeWarning
from sklearn.exceptions import FitFailedWarning

logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", category=OptimizeWarning)
warnings.filterwarnings("ignore", category=FitFailedWarning)


class FuzzyInductor(BaseEstimator, RegressorMixin):
    """FuzzyInductor class."""

    def __init__(self,
                 c=1,
                 k=GaussianKernel(),
                 fuzzifier=ExponentialFuzzifier(), # noqa
                 solver=GurobiSolver(),
                 random_state=None):
        r"""Create an instance of :class:`FuzzyInductor`.

        :param c: Trade-off constant, defaults to 1.
        :type c: `float`
        :param k: Kernel function, defaults to :class:`GaussianKernel()`.
        :type k: :class:`mulearn.kernel.Kernel`
        :param fuzzifier: fuzzifier mapping distance values to membership
           degrees, defaults to `ExponentialFuzzifier()`.
        :type fuzzifier: :class:`mulearn.fuzzifier.Fuzzifier`
        :param solver: Solver to be used to obtain the optimization problem
          solution, defaults to `GurobiSolver()`.
        :type solver: :class:`mulearn.optimization.Solver`
        :param random_state: Seed of the pseudorandom generator.
        :type random_state: `int`
        """
        self.c = c
        self.k = k
        self.fuzzifier = fuzzifier
        self.solver = solver
        self.random_state = random_state
        self.estimated_membership_ = None
        self.x_to_sq_dist_ = None
        self.chis_ = None
        self.gram_ = None
        self.fixed_term_ = None
        self.train_error_ = None

    def __repr__(self, **kwargs):
        return f"FuzzyInductor(c={self.c}, k={self.k}, f={self.fuzzifier}, " \
               f"solver={self.solver})"

    def _fix_object_state(self, X, y):
        """Ensure object consistency."""
        self.X = X
        self.y = y

        def x_to_sq_dist(x_new):
            ret = self.k.compute(x_new, x_new) \
                  - 2 * np.array([self.k.compute(x_i, x_new)
                                  for x_i in X]).dot(self.chis_) \
                  + self.fixed_term_
            return ret

        self.fuzzifier.x_to_sq_dist = x_to_sq_dist

        chi_SV_index = [i for i, (chi, mu) in enumerate(zip(self.chis_, y))
                        if -self.c * (1 - mu) < chi < self.c * mu]

        chi_sq_radius = map(x_to_sq_dist, X[chi_SV_index])
        chi_sq_radius = list(chi_sq_radius)

        if len(chi_sq_radius) == 0:
            self.estimated_membership_ = None
            self.train_error_ = np.inf
            self.chis_ = None
            logger.warning('No support vectors found')
            return self

        self.fuzzifier.sq_radius_05 = np.mean(chi_sq_radius)
        self.fuzzifier.fit(X, y)

        return self.fuzzifier.get_membership()

    def fit(self, X, y, warm_start=False):
        r"""Induce the membership function starting from a labeled sample.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the
          same length
        :param y: Membership for the vectors in `X`.
        :type y: iterable of `float` having the same length of `X`
        :param warm_start: flag triggering the non reinitialization of
          independent variables of the optimization problem, defaults to
          None.
        :type warm_start: `bool`
        :raises: ValueError if the values in `y` are not between 0 and 1, if
          `X` and have different lengths, or if `X` contains elements of
          different lengths.
        :returns: self -- the trained model.
        """
        if type(X) is not np.array:
            X = np.array(X)

        for e in y:
            if e < 0 or e > 1:
                raise ValueError("`y` values should belong to [0, 1]")

        check_X_y(X, y)
        self.random_state = check_random_state(self.random_state)

        if warm_start:
            check_is_fitted(self, ["chis_"])
            if self.chis_ is None:
                raise NotFittedError("chis variable are set to None")
            self.solver.initial_values = self.chis_

        self.chis_ = self.solver.solve(X, y, self.c, self.k)

        if type(self.k) is kernel.PrecomputedKernel:
            idx = X.flatten()
            self.gram_ = self.k.kernel_computations[idx][:, idx]
        else:
            self.gram_ = np.array([[self.k.compute(x1, x2) for x1 in X]
                                   for x2 in X])

        self.fixed_term_ = np.array(self.chis_).dot(self.gram_.dot(self.chis_))

        self.estimated_membership_ = self._fix_object_state(X, y)

        self.train_error_ = np.mean([(self.estimated_membership_(x) - mu) ** 2
                                     for x, mu in zip(X, y)])

        return self

    def decision_function(self, X):
        r"""Compute predictions for the membership function.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the
          same length
        :returns: array of float -- the predictions for each value in `X`.
        """
        check_is_fitted(self, ['chis_', 'estimated_membership_'])
        X = check_array(X)
        return np.array([self.estimated_membership_(x) for x in X]) # noqa

    def predict(self, X, alpha=None):
        r"""Compute predictions for membership to the set.

        Predictions are either computed through the membership function (when
        `alpha` is set to a float in [0, 1]) or obtained via an $\alpha$-cut on
        the same function (when `alpha` is set to `None`).

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :param alpha: $\alpha$-cut value, defaults to `None`.
        :type alpha: float
        :raises: ValueError if `alpha` is set to a value different from
          `None` and not included in $[0, 1]$.
        :returns: array of int -- the predictions for each value in `X`.
        """
        check_is_fitted(self, ['chis_', 'estimated_membership_'])
        X = check_array(X)
        mus = np.array([mu for mu in self.decision_function(X)])
        if alpha is None:
            return mus
        else:
            if alpha < 0 or alpha > 1:
                raise ValueError("alpha cut value should belong to [0, 1]"
                                 f" (provided {alpha})")
            return np.array([1 if mu >= alpha else 0 for mu in mus])

    def score(self, X, y, **kwargs):
        r"""Compute the fuzzifier score.

        Score is obtained as the opposite of MSE between predicted
        membership values and labels.

        :param X: Vectors in data space.
        :type X: iterable of `float` vectors having the same length
        :param y: Labels containing the *gold standard* membership values
          for the vectors in `X`.
        :type y: iterable of `float` having the same length of `X`
        :returns: `float` -- opposite of MSE between the predictions for the
          elements in `X` w.r.t. the labels in `y`.
        """
        check_X_y(X, y)

        if self.estimated_membership_ is None:
            return -np.inf
        else:
            return -np.mean([(self.estimated_membership_(x) - mu) ** 2
                             for x, mu in zip(X, y)])

    def __getstate__(self):
        """Return a serializable description of the fuzzifier."""
        d = copy.deepcopy(self.__dict__)
        del d['estimated_membership_']
        del d['x_to_sq_dist_']
        #print(d)
        return d

    def __setstate__(self, d):
        """Ensure fuzzifier consistency after deserialization."""
        self.__dict__ = d
        try:
            check_is_fitted(self, ['chis_', 'estimated_membership_'])
            self._fix_object_state(self.X, self.y)
            self.__dict__['estimated_membership_'] = self.estimated_membership_
            self.__dict__['x_to_sq_dist_'] = self.x_to_sq_dist_
            self.fuzzifier.x_to_sq_dist = self.x_to_sq_dist_
        except NotFittedError:
            pass



from unittest import TestCase

from mulearn.kernel import *


class TestLinearKernel(TestCase):
    def test_compute(self):
        k = LinearKernel()
        self.assertEqual(k.compute([1, 0, 1], [2, 2, 2]), 4)
        self.assertEqual(k.compute((1, 0, 2), (-1, 2, 5)), 9)
        self.assertAlmostEqual(k.compute([1.2, -0.4, -2], [4, 1.2, .5]), 3.32)
        self.assertAlmostEqual(k.compute((1.2, -0.4, -2), [4, 1.2, .5]), 3.32)
        with self.assertRaises(ValueError):
            k.compute([1, 0, 1], [2, 2])


class TestPolynomialKernel(TestCase):
    def test_compute(self):
        with self.assertRaises(ValueError):
            PolynomialKernel(3.2)

        with self.assertRaises(ValueError):
            PolynomialKernel(-2)

        p = PolynomialKernel(2)
        self.assertEqual(p.compute((1, 0, 2), (-1, 2, 5)), 100)
        self.assertAlmostEqual(p.compute([1.2, -0.4, -2], [4, 1.2, .5]),
                               18.6624)

        p = PolynomialKernel(5)
        self.assertEqual(p.compute((1, 0, 2), [-1, 2, 5]), 10 ** 5)
        self.assertAlmostEqual(p.compute((1.2, -0.4, -2), (4, 1.2, .5)),
                               1504.59195, delta=10**-6)

        with self.assertRaises(ValueError):
            p.compute((1, 0, 2), (-1, 2))


class TestHomogeneousPolynomialKernel(TestCase):
    def test_compute(self):
        with self.assertRaises(ValueError):
            HomogeneousPolynomialKernel(3.2)

        with self.assertRaises(ValueError):
            HomogeneousPolynomialKernel(-2)

        h = HomogeneousPolynomialKernel(2)
        self.assertEqual(h.compute((1, 0, 2), (-1, 2, 5)), 81.0)
        self.assertAlmostEqual(h.compute([1.2, -0.4, -2], [4, 1.2, .5]),
                               11.0224)

        h = HomogeneousPolynomialKernel(5)
        self.assertEqual(h.compute((1, 0, 2), [-1, 2, 5]), 59049.0)
        self.assertAlmostEqual(h.compute((1.2, -0.4, -2), (4, 1.2, .5)),
                               403.357761, delta=10**-6)

        with self.assertRaises(ValueError):
            h.compute((1, 0, 2), (-1, 2)
                      )


class TestGaussianKernel(TestCase):
    def test_compute(self):
        with self.assertRaises(ValueError):
            GaussianKernel(-5)

        k = GaussianKernel(1)
        self.assertAlmostEqual(k.compute((1, 0, 1), (0, 0, 1)), 0.60653065)
        self.assertAlmostEqual(k.compute([-3, 1, 0.5], [1, 1.2, -8]), 6.73e-20)
        self.assertAlmostEqual(k.compute([-1, -4, 3.5], (1, 3.2, 6)), 3.29e-14)

        with self.assertRaises(ValueError):
            k.compute([-1, 3.5], (1, 3.2, 6))


class TestHyperbolicKernel(TestCase):
    def test_compute(self):
        k = HyperbolicKernel(1, 5)
        self.assertAlmostEqual(k.compute((1, 0, 1), (0, 0, 1)), 0.9999877)
        self.assertAlmostEqual(k.compute([-3, 1, 0.5], [1, 1.2, -8]),
                               -0.6640367, delta=10**-7)
        self.assertAlmostEqual(k.compute([-1, -4, 3.5], (1, 3.2, 6)),
                               0.9999999, delta=10**-7)

        with self.assertRaises(ValueError):
            k.compute([-1, 3.5], (1, 3.2, 6))


class TestPrecomputedKernel(TestCase):
    def test_compute(self):
        with self.assertRaises(ValueError):
            PrecomputedKernel(((1, 2), (3, 4, 5)))

        k = PrecomputedKernel(((1, 2), (3, 4)))
        self.assertEqual(k.compute([1], [1]), 4.0)
        self.assertEqual(k.compute([1], [0]), 3.0)

        with self.assertRaises(IndexError):
            k.compute([1], [2])

        with self.assertRaises(TypeError):
            k.compute([0], [1.6])




"""Mulearn distributions.

This module implements the kernel and fuzzifier distributions used in
randomized model selection.
"""

from scipy.stats import uniform
from scipy.stats import rv_continuous

from mulearn.kernel import GaussianKernel
from mulearn.fuzzifier import ExponentialFuzzifier


class GaussianKernelDistribution(rv_continuous):
    """Uniform distribution for gaussian kernels."""

    def __init__(self, low=0, high=1):
        """Build an object of type `GaussianKernelDistribution`.

        :param low: Lower bound of the interval defining the support of the
          uniform distribution associated to the parameter $\sigma$ of a
          :class:`mulearn.kernel.GaussianKernel`, defaults to `0`.
        :type low: `float`
        :param high: Upper bound of the interval defining the support of the
          uniform distribution associated to the parameter $\sigma$ of a
          :class:`mulearn.kernel.GaussianKernel`, defaults to `1`.
        :type high: `float`
        :raises: ValueError if the arguments `low` and `high` do not define
          an interval (that is, `low` is not lower than `high`).
        """
        super().__init__()
        if high <= low:
            raise ValueError(f"the provided upper extreme {high} is lower or"
                             f"equal to the lower one {low}.")
        self.base_dist = uniform(loc=low, scale=high-low)

    def rvs(self, *args, **kwargs):
        """Generate a Gaussian kernel with uniformly distributed parameter.

        :returns: :class:`mulearn.kernel.GaussianKernel` -- Gaussian kernel
          having a parameter uniformly chosen at random in the interval
          having `self.low` and `self.high` as extremes.
        """
        return GaussianKernel(self.base_dist.rvs())


class ExponentialFuzzifierDistribution(rv_continuous):
    """Uniform distribution for gaussian kernels."""

    def __init__(self, low=0, high=1):
        """Build an object of type `ExponentialFuzzifierDistribution`.

        :param low: Lower bound of the interval defining the support of the
          uniform distribution associated to the parameter $\alpha$ of a
          :class:`mulearn.fuzzifier.ExponentialFuzzifier`, defaults to `0`.
        :type low: `float`
        :param high: Upper bound of the interval defining the support of the
          uniform distribution associated to the parameter $\alpha$ of a
          :class:`mulearn.fuzzifier.ExponentialFuzzifier`, defaults to `1`.
        :type high: `float`
        :raises: ValueError if the arguments `low` and `high` do not define
          an interval (that is, `low` is not lower than `high`).
        """
        super().__init__()
        self.base_dist = uniform(loc=low, scale=high-low)

    def rvs(self, *args, **kwargs):
        """Generate an Exponential fuzzifier with uniformly distributed
        parameter.

        :returns: :class:`mulearn.fuzzifier.ExponentialFuzzifier` -- Exponential
          fuzzifier having a parameter uniformly chosen at random in the
          interval having `self.low` and `self.high` as extremes.
        """
        return ExponentialFuzzifier(profile="alpha",
                                    alpha=self.base_dist.rvs())
